"""
Contains global expressions and helper python functions.
"""

# Globals
maven = f"{config['tools']['maven']['command']} {' '.join(config['tools']['maven']['arguments'])}"
java = f"java {' '.join(config['tools']['pantools']['java_arguments'])}"
dataset_name = config["dataset"]["name"]
dataset_type = config["dataset"]["type"]
reference_version = config["tools"]["pantools"]["reference_version"]
validation_directory = f"output/validation/{dataset_name}_{dataset_type}/local-vs-{reference_version}"

def collect_all_cmp(wildcards):
    """
    Collect all cmp file for rule all
    :return: list of cmp files
    """
    input = [f"{validation_directory}/exports/node-properties.csv.cmp",
             f"{validation_directory}/exports/relationship-properties.csv.cmp",
             f"{validation_directory}/exports/sequence-node-anchors.csv.cmp",
             f"{validation_directory}/variation/variation_overview.txt.cmp",
             f"{validation_directory}/gene_classification/gene_classification_overview.cmp",
             f"{validation_directory}/gene_classification/classified_groups.cmp",
             f"{validation_directory}/structure_gene/pangenome_size.txt.cmp",
             f"{validation_directory}/structure_gene/core_accessory_unique_size.csv.cmp",
             f"{validation_directory}/structure_gene/core_dispensable_size.csv.cmp",
             f"{validation_directory}/structure_gene/core_dispensable_total_size.csv.cmp",
             f"{validation_directory}/group/grouping_overview.cmp",
             f"{validation_directory}/group/pantools_homology_groups.cmp",
             f"{validation_directory}/phenotype/phenotype_overview.cmp",
             f"{validation_directory}/functional_classification/core_functions.txt.cmp",
             f"{validation_directory}/functional_classification/accessory_functions.txt.cmp",
             f"{validation_directory}/functional_classification/unique_functions.txt.cmp",
             f"{validation_directory}/function_overview/go_overview.csv.cmp",
             f"{validation_directory}/function_overview/interpro_overview.csv.cmp",
             f"{validation_directory}/function_overview/pfam_overview.csv.cmp",
             f"{validation_directory}/function_overview/tigrfam_overview.csv.cmp",
             f"{validation_directory}/go_enrichment/go_enrichment_overview_per_go.txt.cmp",
             f"{validation_directory}/go_enrichment/go_enrichment.csv.cmp",
             f"{validation_directory}/phylogeny/sites_per_group.csv.cmp",
             f"{validation_directory}/msa_input/genome_order.info.cmp",
             f"{validation_directory}/msa_input/sequences.info.cmp",
             f"{validation_directory}/msa_input/skipped_genomes.info.cmp",
             #f"{validation_directory}/msa_input/trimmed.info.cmp"
             ]
    if dataset_type == "pangenome":
        input.extend([f"{validation_directory}/msa_input/var.fasta.cmp",
                      f"{validation_directory}/msa_input/var_trimmed.fasta.cmp",
                      f"{validation_directory}/msa_output/var.fasta.cmp",
                      f"{validation_directory}/msa_output/var_trimmed.fasta.cmp",
                      f"{validation_directory}/msa_output/var.afa.cmp",
                      f"{validation_directory}/msa_output/var_trimmed.afa.cmp",
                      f"{validation_directory}/msa_input/var.structure.tsv.cmp",
                      f"{validation_directory}/msa_output/variants_trimmed_alignment.info.cmp",
                      f"{validation_directory}/kmer_classification/kmer_classification_overview.cmp",
                      f"{validation_directory}/structure_kmer/pangenome_size_kmer.txt.cmp",
                      f"{validation_directory}/structure_kmer/core_accessory_size.csv.cmp",
                      f"{validation_directory}/structure_kmer/core_accessory_unique_size.csv.cmp",
                      f"{validation_directory}/phylogeny/ANI_distance_matrix.csv.cmp"])
        input.extend(collect_proteins_cmp(wildcards))
        input.extend(collect_sam_cmp(wildcards))
    if dataset_type == "panproteome":
        input.extend([f"{validation_directory}/msa_input/prot.fasta.cmp",
                      f"{validation_directory}/msa_input/prot_trimmed.fasta.cmp",
                      f"{validation_directory}/msa_output/prot.fasta.cmp",
                      f"{validation_directory}/msa_output/prot_trimmed.fasta.cmp",
                      f"{validation_directory}/msa_output/prot.afa.cmp",
                      f"{validation_directory}/msa_output/prot_trimmed.afa.cmp",
                      f"{validation_directory}/msa_output/protein_trimmed_alignment.info.cmp"])
    return input

def collect_proteins_cmp(wildcards):
    """
    Collect all protein comparison files generated from add_annotations output
    :return: list of .cmp files comparing protein files generated by pantools add_annotations
    """
    checkpoint_output = checkpoints.add_annotations.get(
        dataset_name=dataset_name,
        dataset_type=dataset_type,
        build_version=reference_version
    ).output[0]
    checkpoints.add_annotations.get(
        dataset_name=dataset_name,
        dataset_type=dataset_type,
        build_version="local"
    )
    filenames=glob_wildcards(os.path.join(checkpoint_output, "{filename}.fasta")).filename
    return expand(f"{validation_directory}/proteins/{{filename}}.cmp", filename=filenames)

def collect_sam_cmp(wildcards):
    """
    Return paths to all files produced by the checkpoint 'map_reads'
    SAM files comparison .TSV files. A .cmp file will be generated for each pair of SAM
    files, a pair being the alignments against genome x for the local version and the
    reference version.
    :return: list of .cmp files comparing SAM files generated by pantools map
    """
    checkpoint_output = checkpoints.map_reads.get(
        dataset_name=dataset_name,
        dataset_type=dataset_type,
        build_version=reference_version
    ).output[0]
    checkpoints.map_reads.get(
        dataset_name=dataset_name,
        dataset_type=dataset_type,
        build_version="local"
    )
    filenames=glob_wildcards(os.path.join(checkpoint_output, "{filename}.sam")).filename
    return expand(f"{validation_directory}/sams/{{filename}}.sam.cmp", filename=filenames)

def collect_aggregate_msa_input(wildcards):
    """
    Collect files generated by msa for aggregation.
    :param wildcards: wildcards of the aggregation rule
    :return: list of msa files
    """
    if dataset_type == "pangenome":
        checkpoint_output = checkpoints.msa_var.get(**wildcards).output[1]
    else:
        checkpoint_output = checkpoints.msa_prot.get(**wildcards).output[1]
    filename = f"{{id}}/{wildcards.subdir}/{wildcards.filename}.{wildcards.ext}"
    id = glob_wildcards(os.path.join(checkpoint_output, filename)).id
    return expand(f"{checkpoint_output}/{filename}", id=id)